@@ -76,137 +76,91 @@ static int uhci_td_result(struct uhci_device *dev, struct uhci_td *td)
 	return status;		
 }
 
-#if 0
-
 /*
- * Functions for manipulating tds and q contexts
+ * Inserts a td into qh list at the top.
+ *
+ * Careful about atomicity: even on UP this
+ * requires a locked access due to the concurrent
+ * DMA engine.
+ *
+ * NOTE! This assumes that first->last is a valid
+ * list of TD's with the proper backpointers set
+ * up and all..
  */
-
-/*
- * Convert a UHCI link pointer to a uhci_qh pointer if it is a qh. Return
- * NULL otherwise (e.g. td or terminate).
- */
-static struct uhci_qh *uhci_link_to_qh(unsigned int element)
+static void uhci_insert_tds_in_qh(struct uhci_qh *qh, struct uhci_td *first, struct uhci_td *last)
 {
-	/* if element==terminate or element points to a td return NULL */
-	if((element & 1) || !(element & 2)) {
-		return NULL;
+	unsigned int link = qh->element;
+	unsigned int new = 4 | virt_to_bus(first);
+
+	for (;;) {
+		unsigned char success;
+
+		last->link = link;
+		first->backptr = &qh->element;
+		asm volatile("lock ; cmpxchg %4,%2 ; sete %0"
+			:"=q" (success), "=a" (link)
+			:"m" (qh->element), "1" (link), "r" (new)
+			:"memory");
+		if (success) {
+			/* Was there a successor entry? Fix it's backpointer.. */
+			if ((link & 1) == 0) {
+				struct uhci_td *next = bus_to_virt(link & ~15);
+				next->backptr = &last->link;
+			}
+			break;
+		}
 	}
-	return (struct uhci_qh *) bus_to_virt(element & ~0xf);
 }
 
-#endif
-
-/*
- * Convert a UHCI link pointer to a uhci_td pointer if it is a td. Return
- * NULL otherwise (e.g. qh or terminate).
- */
-struct uhci_td *uhci_link_to_td(unsigned int element)
+static inline void uhci_insert_td_in_qh(struct uhci_qh *qh, struct uhci_td *td)
 {
-	/* if element==terminate or element points to a qh return NULL */
-	if(element & 3) {
-		return NULL;
-	}
-	return (struct uhci_td *) bus_to_virt(element & ~0xf);
-}
-
-/* 
- * Searches qh for td. Returns virt pointer to the link that points to td if 
- * found, NULL if not found. If td==NULL, returns pointer to last link in the
- * qh.
- */
-static unsigned int *uhci_find_td_in_qh(struct uhci_qh *qh, struct uhci_td *td)
-{
-	struct uhci_td *child_td, *parent_td;
-
-	if(uhci_link_to_td(qh->element) == td)
-		return &qh->element;
-
-	for(parent_td = NULL, child_td = uhci_link_to_td(qh->element);
-	    child_td != NULL; 
-	    parent_td = child_td, child_td = uhci_link_to_td(child_td->link)) {
-		if(child_td == td)
-			return &parent_td->link;
-	}
-	/* HACK: fix the loop instead, later */
-	if(child_td == td)
-		return &parent_td->link;
-
-	return NULL;
-}
-
-/* 
- * Insert the td list [first_td ... last_td] into qh (at the tail, of course).
- */
-/* FIXME: if we keep multiple qhs linked vertically, this will unlink all but
-   the first one, lucky thing we don't do that :) */
-static int uhci_insert_tds_in_qh(struct uhci_qh *qh, struct uhci_td *first_td,
-				 struct uhci_td *last_td)
-{
-	unsigned int *link;
-
-	if(uhci_find_td_in_qh(qh, first_td) != NULL ||
-	   uhci_find_td_in_qh(qh, last_td) != NULL) {
-		printk("Refusing to insert tds %p thru %p into qh %p, tds already inserted\n", first_td, last_td, qh);
-		return -1;
-	}
-
-	/* it would be more efficient to keep the tail pointer around, but
-	   let's get this thing working first */
-	link = uhci_find_td_in_qh(qh, NULL);
-	if(link == NULL) {
-		printk("Can't find end of qh %p\n", qh);
-		return -1;
-	}
-
-	/* is the depth first flag necessary? */
-	last_td->link = (*link == 1) ? 1 : (4 | *link);
-	wmb();
-	*link = virt_to_bus(first_td);
-	return 0;
-}
-
-/*
- * Inserts td into qh.
- */
-static int uhci_insert_td_in_qh(struct uhci_qh *qh, struct uhci_td *td)
-{
-	return uhci_insert_tds_in_qh(qh, td, td);
+	uhci_insert_tds_in_qh(qh, td, td);
 }
 
 /*
  * Removes td from qh if present.
+ *
+ * NOTE! We keep track of both forward and back-pointers,
+ * so this should be trivial, right?
+ *
+ * Wrong. While all TD insert/remove operations are synchronous
+ * on the CPU, the UHCI controller can (and does) play with the
+ * very first forward pointer. So we need to validate the backptr
+ * before we change it, so that we don't by mistake reset the QH
+ * head to something old.
  */
-static void uhci_remove_td_from_qh(struct uhci_qh *qh, struct uhci_td *td)
+static void uhci_remove_td(struct uhci_td *td)
 {
-	unsigned int *link = uhci_find_td_in_qh(qh, td);
-	if(link != NULL) {
-		/* we need to remove it then */
-		*link = td->link;
-	}
-}
+	unsigned int *backptr = td->backptr;
+	unsigned int link = td->link;
+	unsigned int me;
 
-/*
- * Removes the td list [first_td ... last_td] from qh if present. Complains
- * if only one end of the list can be found.
- */
-static void uhci_remove_tds_from_qh(struct uhci_qh *qh, 
-				    struct uhci_td *first_td,
-				    struct uhci_td *last_td)
-{
-	unsigned int *first_link, *last_link;
-
-	first_link = uhci_find_td_in_qh(qh, first_td);
-	last_link = uhci_find_td_in_qh(qh, last_td);
-
-	if(first_link == NULL && last_link == NULL) {
-		return; /* nothing to do */
-	} else if(first_link == NULL || last_link == NULL) {
-		printk("What is this? Partially linked td list! [%p, %p] at (%p, %p)\n", first_td, last_td, first_link, last_link);
+	if (!backptr)
 		return;
+
+	td->backptr = NULL;
+
+	/*
+	 * This is the easy case: the UHCI will never change "td->link",
+	 * so we can always just look at that and fix up the backpointer
+	 * of any next element..
+	 */
+	if (!(link & 1)) {
+		struct uhci_td *next = bus_to_virt(link & ~15);
+		next->backptr = backptr;
 	}
 
-	*first_link = last_td->link;
+	/*
+	 * The nasty case is "backptr->next", which we need to
+	 * update to "link" _only_ if "backptr" still points
+	 * to us (it may not: maybe backptr is a QH->element
+	 * pointer and the UHCI has changed the value).
+	 */
+	me = virt_to_bus(td) | (0xe & *backptr);
+	asm volatile("lock ; cmpxchg %0,%1"
+		:
+		:"r" (link), "m" (*backptr), "a" (me)
+		:"memory");
 }
 
 /*
@@ -214,12 +168,11 @@ static void uhci_remove_tds_from_qh(struct uhci_qh *qh,
  */
 static spinlock_t irqlist_lock = SPIN_LOCK_UNLOCKED;
 
-static void uhci_add_irq_list(struct uhci *uhci, struct uhci_td *td, usb_device_irq completed, void *data, void *dev_id)
+static void uhci_add_irq_list(struct uhci *uhci, struct uhci_td *td, usb_device_irq completed, void *dev_id)
 {
 	unsigned long flags;
 
 	td->completed = completed;
-	td->data = data;
 	td->dev_id = dev_id;
 
 	spin_lock_irqsave(&irqlist_lock, flags);
@@ -239,14 +192,10 @@ static void uhci_remove_irq_list(struct uhci_td *td)
 /*
  * Request a interrupt handler..
  */
-static int uhci_request_irq(struct usb_device * usb_dev, unsigned int pipe,
-	usb_device_irq handler, void *data, int period, void *dev_id)
+static int uhci_request_irq(struct usb_device * usb_dev, unsigned int pipe, usb_device_irq handler, int period, void *dev_id)
 {
 	struct uhci_device *dev = usb_to_uhci(usb_dev);
-	/* FIXME: sending a control message to this device after interrupt is
-	   allocated will destroy the interrupt td. -keryan */
-	/* cheesy hack */
-	struct uhci_td *td = dev->td + 30;
+	struct uhci_td *td = dev->td + 30;					/* cheesy hack */
 	struct uhci_qh *interrupt_qh = &dev->uhci->root_hub->interrupt8_qh;	/* Ignore the period for now */
 
 	unsigned int destination, status;
@@ -266,7 +215,7 @@ static int uhci_request_irq(struct usb_device * usb_dev, unsigned int pipe,
 	td->info = destination | (7 << 21);	/* 8 bytes of data */
 	td->buffer = virt_to_bus(dev->data);
 
-	uhci_add_irq_list(dev->uhci, td, handler, data, dev_id);
+	uhci_add_irq_list(dev->uhci, td, handler, dev_id);
 
 	uhci_insert_td_in_qh(interrupt_qh, td);
 	return 0;
@@ -282,7 +231,7 @@ static int uhci_request_irq(struct usb_device * usb_dev, unsigned int pipe,
  */
 static struct wait_queue *control_wakeup;
 
-static int uhci_control_completed(int status, void *data, void *buffer, void *dev_id)
+static int uhci_control_completed(int status, void *buffer, void *dev_id)
 {
 	wake_up(&control_wakeup);
 	return 0;			/* Don't re-instate */
@@ -297,7 +246,7 @@ static int uhci_run_control(struct uhci_device *dev, struct uhci_td *td)
 	current->state = TASK_UNINTERRUPTIBLE;
 	add_wait_queue(&control_wakeup, &wait);
 
-	uhci_add_irq_list(dev->uhci, td, uhci_control_completed, NULL, NULL);
+	uhci_add_irq_list(dev->uhci, td, uhci_control_completed, NULL);
 	
 	ctrl_qh = &dev->uhci->root_hub->control_qh;
 	
@@ -316,10 +265,15 @@ static int uhci_run_control(struct uhci_device *dev, struct uhci_td *td)
 #if 0
 	printk("Looking for tds [%p, %p]\n", dev->control_td, td);
 #endif
-	/* remove whole list */
-	uhci_remove_tds_from_qh(ctrl_qh, dev->control_td, td);
-
-	/*	dev->uhci->root_hub->control_qh.element = 1; */
+	{
+		int maxcount = 100;
+		struct uhci_td *first = dev->control_td;
+		do {
+			uhci_remove_td(first);
+			first++;
+			if (!--maxcount) break;
+		} while (first <= td);
+	}
 
 	return uhci_td_result(dev, td);
 }
@@ -345,8 +299,6 @@ static int uhci_run_control(struct uhci_device *dev, struct uhci_td *td)
  * information, that's just ridiculously high. Most
  * control messages have just a few bytes of data.
  */
-/* FIXME: Sending control messages to a device that has an interrupt will 
-   destroy the interrupt td. -keryan */
 static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void *cmd, void *data, int len)
 {
 	struct uhci_device *dev = usb_to_uhci(usb_dev);
@@ -395,6 +347,7 @@ static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void
 		td->status = status;					/* Status */
 		td->info = destination | ((pktsze-1) << 21);		/* pktsze bytes of data */
 		td->buffer = virt_to_bus(data);
+		td->backptr = &td[-1].link;
 
 		td++;
 		data += maxsze;
@@ -411,6 +364,7 @@ static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void
 	td->status = status | (1 << 24);		/* IOC */
 	td->info = destination | (0x7ff << 21);		/* 0 bytes of data */
 	td->buffer = 0;
+	td->backptr = &td[-1].link;
 
 	/* Start it up.. */
 	return uhci_run_control(dev, td);
@@ -461,7 +415,7 @@ static int uhci_usb_deallocate(struct usb_device *usb_dev)
 
 		/* Remove the td from any queues */
 		for(j = 0; j < UHCI_MAXQH; ++j)
-			uhci_remove_td_from_qh(dev->qh + j, td);
+			uhci_remove_td(td);
 		
 		/* And remove it from the irq list, if it's active */
 		if(td->status & (1 << 23))
@@ -606,13 +560,13 @@ static void uhci_interrupt_notify(struct uhci *uhci)
 			/* remove from IRQ list */
 			__list_del(tmp->prev, next);
 			INIT_LIST_HEAD(tmp);
-			if (td->completed(td->status, td->data, bus_to_virt(td->buffer), td->dev_id)) {
+			if (td->completed(td->status, bus_to_virt(td->buffer), td->dev_id)) {
 				struct uhci_qh * interrupt_qh = &uhci->root_hub->interrupt8_qh;
 				list_add(&td->irq_list, &uhci->interrupt_list);
 				td->info ^= 1 << 19; /* toggle between data0 and data1 */
 				td->status = (td->status & 0x2f000000) | (1 << 23) | (1 << 24);	/* active */
-				/* FIXME: doesn't hc remove this td? */
-				uhci_remove_td_from_qh(interrupt_qh, td);
+
+				uhci_remove_td(td);
 				uhci_insert_td_in_qh(interrupt_qh, td);
 			}
 		}
